
import os
import glob

from IPython import embed

import numpy
from matplotlib import pyplot

from astropy.io import fits

from mangadap.dapfits import DAPMapsBitMask, DAPQualityBitMask
from mangadap.config import defaults
from mangadap.util.datatable import DataTable
from mangadap.util.bitmask import BitMask
from mangadap.util.fileio import channel_dictionary
from mangadap.util.fitsutil import DAPFitsUtil

class HalphaBitMask(BitMask):
    def __init__(self):
        super(HalphaBitMask, self).__init__(['CRITCUBE', 'BADFLUX', 'BADEW', 'BADVEL', 'BADSIG'])


class HalphaDataTable(DataTable):
    def __init__(self, shape=None):
        bm = HalphaBitMask()
        _float = numpy.float32
        _int = numpy.int32
        datamodel = dict(PLATE=dict(typ=_int), IFU=dict(typ=_int), X=dict(typ=_int),
                         Y=dict(typ=_int), BINID=dict(typ=_int), R=dict(typ=_float),
                         R_RE=dict(typ=_float), R_KPC=dict(typ=_float), GCNT=dict(typ=_float),
                         GSNR=dict(typ=_float), FLUX=dict(typ=_float), FLUX_ERR=dict(typ=_float),
                         AMP=dict(typ=_float), ANR=dict(typ=_float), RCHI2=dict(typ=_float),
                         CNT=dict(typ=_float), EW=dict(typ=_float), EW_ERR=dict(typ=_float),
                         VEL=dict(typ=_float), VEL_ERR=dict(typ=_float), SIG=dict(typ=_float),
                         SIG_ERR=dict(typ=_float), SIG_INST=dict(typ=_float),
                         MASK=dict(typ=bm.minimum_dtype()))
        keys = list(datamodel.keys())
        super(HalphaDataTable, self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                                              shape=shape)


def get_halpha_data(plt, ifu, hdu):
    el = channel_dictionary(hdu, 'EMLINE_GFLUX')
    uniq_bins, uniq_indx, uniq_cnt = map(lambda x : x[1:],
                                         numpy.unique(hdu['BINID'].data[3].ravel(),
                                                      return_index=True, return_counts=True))
    keep = numpy.logical_not(hdu['EMLINE_GFLUX_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['EMLINE_GEW_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['EMLINE_GVEL_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['EMLINE_GSIGMA_MASK'].data.ravel()[uniq_indx] > 0)
    uniq_bins = uniq_bins[keep]
    uniq_indx = uniq_indx[keep]
    uniq_cnt = uniq_cnt[keep]
    data = HalphaDataTable(shape=uniq_bins.shape)

    data['PLATE'] = plt
    data['IFU'] = ifu
    data['Y'], data['X'] = numpy.unravel_index(uniq_indx, hdu['BINID'].shape[1:])
    data['BINID'] = uniq_bins
    data['R'] = hdu['SPX_ELLCOO'].data[0].ravel()[uniq_indx]
    data['R_RE'] = hdu['SPX_ELLCOO'].data[1].ravel()[uniq_indx]
    data['R_KPC'] = hdu['SPX_ELLCOO'].data[2].ravel()[uniq_indx]
    data['GCNT'] = hdu['SPX_MFLUX'].data.ravel()[uniq_indx]
    data['GSNR'] = hdu['SPX_SNR'].data.ravel()[uniq_indx]
    data['FLUX'] = hdu['EMLINE_GFLUX'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['FLUX_ERR'] = numpy.ma.power(
                            hdu['EMLINE_GFLUX_IVAR'].data[el['Ha-6564']].ravel()[uniq_indx],
                            -0.5).filled(-1)
    data['AMP'] = hdu['EMLINE_GA'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['ANR'] = hdu['EMLINE_GANR'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['RCHI2'] = hdu['EMLINE_LFOM'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['CNT'] = hdu['EMLINE_GEW_CNT'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['EW'] = hdu['EMLINE_GEW'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['EW_ERR'] = numpy.ma.power(hdu['EMLINE_GEW_IVAR'].data[el['Ha-6564']].ravel()[uniq_indx],
                                    -0.5).filled(-1)
    data['VEL'] = hdu['EMLINE_GVEL'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['VEL_ERR'] = numpy.ma.power(hdu['EMLINE_GVEL_IVAR'].data[el['Ha-6564']].ravel()[uniq_indx],
                                    -0.5).filled(-1)
    data['SIG'] = hdu['EMLINE_GSIGMA'].data[el['Ha-6564']].ravel()[uniq_indx]
    data['SIG_ERR'] = numpy.ma.power(
                            hdu['EMLINE_GSIGMA_IVAR'].data[el['Ha-6564']].ravel()[uniq_indx],
                            -0.5).filled(-1)
    data['SIG_INST'] = hdu['EMLINE_INSTSIGMA'].data[el['Ha-6564']].ravel()[uniq_indx]
   
    # Deal with the masking
    qual_bm = DAPQualityBitMask()
    data_bm = HalphaBitMask()
    if qual_bm.flagged(hdu[0].header['DAPQUAL'], flag='CRITICAL'):
        data['MASK'] = data_bm.turn_on(data['MASK'], 'CRITCUBE')
    indx = hdu['EMLINE_GFLUX_MASK'].data[el['Ha-6564']].ravel()[uniq_indx] > 0
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADFLUX')
    indx = hdu['EMLINE_GEW_MASK'].data[el['Ha-6564']].ravel()[uniq_indx] > 0
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADEW')
    indx = (hdu['EMLINE_GVEL_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['EMLINE_GVEL_IVAR'].data.ravel()[uniq_indx] > 0)
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADVEL')
    indx = (hdu['EMLINE_GSIGMA_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['EMLINE_GSIGMA_IVAR'].data.ravel()[uniq_indx] > 0)
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADSIG')

    return data


class StellarBitMask(BitMask):
    def __init__(self):
        super(StellarBitMask, self).__init__(['CRITCUBE', 'BADFLUX', 'BADVEL', 'BADSIG'])


class StellarDataTable(DataTable):
    def __init__(self, shape=None):
        bm = StellarBitMask()
        _float = numpy.float32
        _int = numpy.int32
        datamodel = dict(PLATE=dict(typ=_int), IFU=dict(typ=_int), X=dict(typ=_int),
                         Y=dict(typ=_int), BINID=dict(typ=_int), NBIN=dict(typ=_int),
                         R=dict(typ=_float), R_RE=dict(typ=_float), R_KPC=dict(typ=_float),
                         GCNT=dict(typ=_float), GSNR=dict(typ=_float), RCHI2=dict(typ=_float),
                         VEL=dict(typ=_float), VEL_ERR=dict(typ=_float), SIG=dict(typ=_float),
                         SIG_ERR=dict(typ=_float), SIG_CORR=dict(typ=_float),
                         MASK=dict(typ=bm.minimum_dtype()))
        keys = list(datamodel.keys())
        super(StellarDataTable, self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                                               shape=shape)


def get_stellar_data(plt, ifu, hdu):
    uniq_bins, uniq_indx, uniq_cnt = map(lambda x : x[1:],
                                         numpy.unique(hdu['BINID'].data[1].ravel(),
                                                      return_index=True, return_counts=True))
    keep = numpy.logical_not(hdu['BIN_MFLUX_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['STELLAR_VEL_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['STELLAR_SIGMA_MASK'].data.ravel()[uniq_indx] > 0)
    uniq_bins = uniq_bins[keep]
    uniq_indx = uniq_indx[keep]
    uniq_cnt = uniq_cnt[keep]
    data = StellarDataTable(shape=uniq_bins.shape)

    data['PLATE'] = plt
    data['IFU'] = ifu
    data['Y'], data['X'] = numpy.unravel_index(uniq_indx, hdu['BINID'].shape[1:])
    data['BINID'] = uniq_bins
    data['NBIN'] = uniq_cnt
    data['R'] = hdu['BIN_LWELLCOO'].data[0].ravel()[uniq_indx]
    data['R_RE'] = hdu['BIN_LWELLCOO'].data[1].ravel()[uniq_indx]
    data['R_KPC'] = hdu['BIN_LWELLCOO'].data[2].ravel()[uniq_indx]
    data['GCNT'] = hdu['BIN_MFLUX'].data.ravel()[uniq_indx]
    data['GSNR'] = hdu['BIN_SNR'].data.ravel()[uniq_indx]
    data['RCHI2'] = hdu['STELLAR_FOM'].data[2].ravel()[uniq_indx]
    data['VEL'] = hdu['STELLAR_VEL'].data.ravel()[uniq_indx]
    data['VEL_ERR'] = numpy.ma.power(hdu['STELLAR_VEL_IVAR'].data.ravel()[uniq_indx],
                                     -0.5).filled(-1)
    data['SIG'] = hdu['STELLAR_SIGMA'].data.ravel()[uniq_indx]
    data['SIG_ERR'] = numpy.ma.power(hdu['STELLAR_SIGMA_IVAR'].data.ravel()[uniq_indx],
                                     -0.5).filled(-1)
    data['SIG_CORR'] = hdu['STELLAR_SIGMACORR'].data.ravel()[uniq_indx]
   
    # Deal with the masking
    # maps_bm = DAPMapsBitMask()
    qual_bm = DAPQualityBitMask()
    data_bm = StellarBitMask()
    if qual_bm.flagged(hdu[0].header['DAPQUAL'], flag='CRITICAL'):
        data['MASK'] = data_bm.turn_on(data['MASK'], 'CRITCUBE')
    indx = hdu['BIN_MFLUX_MASK'].data.ravel()[uniq_indx] > 0
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADFLUX')
    indx = (hdu['STELLAR_VEL_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['STELLAR_VEL_IVAR'].data.ravel()[uniq_indx] > 0)
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADVEL')
    indx = (hdu['STELLAR_SIGMA_MASK'].data.ravel()[uniq_indx] > 0) \
            | numpy.logical_not(hdu['STELLAR_SIGMA_IVAR'].data.ravel()[uniq_indx] > 0)
    if numpy.any(indx):
        data['MASK'][indx] = data_bm.turn_on(data['MASK'][indx], 'BADSIG')

    return data


def get_maps_files(daptype='HYB10-MILESHC-MASTARHC2'):
#    return ['/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/12704/manga-7443-12704-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/12701/manga-7443-12701-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/12703/manga-7443-12703-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/12705/manga-7443-12705-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/1901/manga-7443-1901-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/1902/manga-7443-1902-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz', '/uufs/chpc.utah.edu/common/home/sdss/mangawork/manga/spectro/analysis/v3_0_1/3.0.1/HYB10-MILESHC-MASTARHC2/7443/12702/manga-7443-12702-MAPS-HYB10-MILESHC-MASTARHC2.fits.gz']

    root = os.path.join(defaults.dap_analysis_path(dapver='3.0.1'), daptype)
    return glob.glob(os.path.join(root, '*', '*', '*MAPS*fits.gz'))
    

def main():
    maps_files = get_maps_files()
    nmaps = len(maps_files)
    nbin = 100

    j = 0
    stellar_data = StellarDataTable()
    halpha_data = HalphaDataTable()
    stellar_root = 'stellar_data_mpl10'
    halpha_root = 'halpha_data_mpl10'
    for i in range(nmaps):
        plt, ifu = map(lambda x : int(x), os.path.split(maps_files[i])[1].split('-')[1:3])
        with fits.open(maps_files[i]) as hdu:
            stellar_data.append(get_stellar_data(plt, ifu, hdu))
            halpha_data.append(get_halpha_data(plt, ifu, hdu))
        print('{0}/{1} : {2} : {3}'.format(i+1, nmaps, stellar_data.size, halpha_data.size),
              end='\r')
        j += 1
        if j % nbin == 0 or i == nmaps-1:
            k = i // nbin
            #print(nmaps, nbin, i, j, k)
            ofile = '{0}_{1}.fits.gz'.format(stellar_root, str(k).zfill(3))
            DAPFitsUtil.write(fits.HDUList([fits.PrimaryHDU(),
                                            stellar_data.to_hdu(name='STELLAR')]),
                              ofile, clobber=True)
            ofile = '{0}_{1}.fits.gz'.format(halpha_root, str(k).zfill(3))
            DAPFitsUtil.write(fits.HDUList([fits.PrimaryHDU(),
                                            halpha_data.to_hdu(name='HALPHA')]),
                              ofile, clobber=True)
            stellar_data = StellarDataTable()
            halpha_data = HalphaDataTable()
    print('{0}/{0} : {1} : {2}'.format(nmaps, stellar_data.size, halpha_data.size))

#    ofile = 'stellar_data_mpl10.fits.gz'
#    DAPFitsUtil.write(fits.HDUList([fits.PrimaryHDU(), stellar_data.to_hdu(name='STELLAR')]),
#                      ofile, clobber=True)
#
#    ofile = 'halpha_data_mpl10.fits.gz'
#    DAPFitsUtil.write(fits.HDUList([fits.PrimaryHDU(), halpha_data.to_hdu(name='HALPHA')]),
#                      ofile, clobber=True)
   
if __name__ == '__main__':
    main()




from pathlib import Path
from IPython import embed

from mangadap.tests.util import data_test_file
from mangadap.config.analysisplan import AnalysisPlan
from mangadap.config.manga import MaNGAAnalysisPlan, MaNGAConfig


def test_default():
    plan = AnalysisPlan.default()
    assert len(plan.keys()) == 1, 'Default is to perform one plan.'
    assert list(plan.keys())[0] == 'default', 'Name changed'

    assert plan.rdxqa['default']['key'] == 'SNRG', 'Default DRP reduction QA key changed.'
    assert plan.elfit['default']['key'] == 'EFITHC2', 'Default emission-line fit key changed.'
    assert plan.common_path() == Path('.').resolve() / 'common', 'Common path changed'


def test_read():
    plan = AnalysisPlan.from_toml(data_test_file('global_bin.toml'))

    assert len(plan.keys()) == 1, 'Number of example plans changed'
    assert list(plan.keys())[0] == 'default', 'Name changed'

    assert plan.binning['default']['key'] == 'ALL', 'Default DRP reduction QA key changed.'
    assert plan.elfit['default']['key'] == 'EFITSSP', 'Default emission-line fit key changed.'


def test_multi_read():
    plan = AnalysisPlan.from_toml(data_test_file('dr17.toml'))

    assert len(plan.keys()) == 4, 'Number of example plans changed'

    assert plan['plan3']['key'] == 'HYB10-MILESHC-MASTARSSP', 'Plan key changed'
    assert plan.binning['plan1']['key'] == 'SPX', 'Binning changed'
    assert plan.elfit['plan4']['key'] == 'EFITHC2DB', 'Emission-line fitting key changed'


def test_mangaplan():
    cfg = MaNGAConfig(7815, 3702)
    plan = MaNGAAnalysisPlan.from_toml(data_test_file('dr17.toml'), cube=cfg)

    assert plan.dap_file_root(cfg) == 'manga-7815-3702', 'Bad root'
    assert plan.dap_file_root(cfg, mode='MAPS', plan_index=0) \
                == 'manga-7815-3702-MAPS-SPX-MILESHC-MASTARSSP', 'Bad full root'
    assert plan.common_path().parts[-2:] == (str(cfg.plate), str(cfg.ifudesign)), \
                'Bad common subdirectories'
    assert plan.method_path().parts[-2:] == (str(cfg.plate), str(cfg.ifudesign)), \
                'Bad method subdirectories'
    assert plan.method_path(qa=True).parts[-2:] == (str(cfg.ifudesign), 'qa'), \
                'Bad qa subdirectories'





#!/usr/bin/env python3

import os
import time
import numpy

from IPython import embed

from astropy.io import fits
from matplotlib import pyplot
from matplotlib.ticker import NullFormatter

from mangadap.config import defaults
from mangadap.config import manga
from mangadap.datacube import MaNGADataCube
from mangadap.util.constants import DAPConstants
from mangadap.util.fitsutil import DAPFitsUtil
from mangadap.util.resolution import SpectralResolution
from mangadap.util.filter import interpolate_masked_vector
from mangadap.util.fileio import read_template_spectrum
from mangadap.util.pixelmask import SpectralPixelMask

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionlinedb import EmissionLineDB

from mangadap.proc.templatelibrary import available_template_libraries
from mangadap.proc.templatelibrary import TemplateLibraryDef, TemplateLibrary
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.sasuke import Sasuke
from mangadap.proc.ppxffit import PPXFFit
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel, StellarContinuumModelBitMask
from mangadap.proc.emissionlinemodel import EmissionLineModelBitMask

#-----------------------------------------------------------------------------
def get_redshift(plt, ifu, drpall_file=None):
    """
    Get the redshift of a galaxy from the DRPall file.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        drapall_file (:obj:`str`, optional):
            DRPall file. If None, attempts to use the default path to
            the file using environmental variables.
    
    Returns:
        :obj:`float`: The redshift to the galaxy observed by the
        provided PLATEIFU.
    """
    if drpall_file is None:
        drpall_file = manga.drpall_file()
    if not drpall_file.exists():
        raise FileNotFoundError(f'Could not find DRPall file: {drpall_file}')
    hdu = fits.open(str(drpall_file))
    indx = hdu[1].data['PLATEIFU'] == '{0}-{1}'.format(plt, ifu)
    return hdu[1].data['NSA_Z'][indx][0]


def get_spectra(plt, ifu, x, y, directory_path=None):
    """
    Extract spectra from a MaNGA observation.

    Args:
        plt (:obj:`int`):
            Plate number
        ifu (:obj:`int`):
            IFU identifier
        x (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the RA axis.
        y (:obj:`int`, `numpy.ndarray`_):
            The spaxel coordinate along the DEC axis.
        directory_path (:obj:`str`, optional):
            Directory with the DRP LOGCUBE file. If None, uses the
            default directory path based on the environmental
            variables.

    Returns:
        :obj:`tuple`: Returns 4 numpy vectors: The wavelength, flux,
        flux inverse variance, and spectral resolution extracted from
        the datacube.
    """
    cube = MaNGADataCube.from_plateifu(plt, ifu, directory_path=directory_path)
    flat_indx = cube.spatial_shape[1]*x+y
    # This function always returns as masked array
    flux = cube.copy_to_masked_array(attr='flux', flag=cube.do_not_fit_flags())
    ivar = cube.copy_to_masked_array(attr='ivar', flag=cube.do_not_fit_flags())
    sres = cube.copy_to_array(attr='sres')
    return cube.wave, flux[flat_indx,:], ivar[flat_indx,:], sres[flat_indx,:]


def create_mastar_template_spectrum():
    output_file = 'mastar-8047-9101-tpl.fits'

    hdu = fits.open('mastar-LOG-8047-9101.fits.gz')
    wave = hdu['MASTAR'].data['WAVE'].ravel()
    flux = hdu['MASTAR'].data['FLUX'].ravel()
    sres = interpolate_masked_vector(numpy.ma.power(DAPConstants.sig2fwhm
                                            * hdu['MASTAR'].data['PREDISP'].ravel() / wave, -1))

    hdr = fits.Header()
    hdr['CRPIX1'] = 1
    hdr['CRVAL1'] = numpy.log10(wave[0])
    hdr['CDELT1'] = numpy.diff(numpy.log10(wave[:2]))[0]
    hdr['CD1_1'] = hdr['CDELT1']
    fits.HDUList([ fits.PrimaryHDU(flux, header=hdr),
                   fits.ImageHDU(sres, name='SPECRES')]).writeto(output_file, overwrite=True)
    
        

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.perf_counter()

    # Get the list of available libraries
    tpllib_list = available_template_libraries()
#    print([tpl['key'] for tpl in tpllib_list])
#    print(tpllib_list[0].data)

    # ------------------------------------------------------------------
    # Create a library based on a subset of an existing one

    # Add a "new" library that only has 10 Gyr SSPs from MIUSCAT
    file_search = str(defaults.dap_source_dir() / 'data' / 'spectral_templates' / 'miuscat'
                        /  '*T10.0000.fits')

    # Define the template library
    tpllib_list += [ TemplateLibraryDef('OLD',
                                        file_search=file_search,
                                        fwhm=2.51,
                                        in_vacuum=False,
                                        wave_limit=numpy.array([3480, 9430]),
                                        log10=False) ]

    # ------------------------------------------------------------------
    # Create a new library based on a MaStar spectrum
    if not os.path.isfile('mastar-8047-9101-tpl.fits'):
        create_mastar_template_spectrum()
    tpllib_list += [ TemplateLibraryDef('MaStar',
                                        file_search='mastar-8047-9101-tpl.fits',
                                        sres_ext='SPECRES',
                                        in_vacuum=True,
                                        log10=True) ]

    # ------------------------------------------------------------------
    # Let's try to use them!

    # Plate-IFU to use
    plt = 7815
    ifu = 3702
    # Spaxel coordinates
    x = 21
    y = 21

    # Template keywords
#    sc_tpl_key = 'MILESHC'
#    sc_tpl_key = 'OLD'
    sc_tpl_key = 'MaStar'

    # Template pixel scale a factor of 4 smaller than galaxy data
    velscale_ratio = 4

    # Get the redshift
    drpver = 'v3_1_1'
    directory_path = defaults.dap_source_dir() / 'data' / 'remote'
    drpall_file = directory_path / f'drpall-{drpver}.fits'
    z = numpy.array([get_redshift(plt, ifu, drpall_file)])
    print('Redshift: {0}'.format(z[0]))
    dispersion = numpy.array([100.])

    # Read a spectrum
    print('reading spectrum')
    wave, flux, ivar, sres = get_spectra(plt, ifu, x, y, directory_path=directory_path)

    # Fitting functions expect data to be in 2D arrays (for now):
    flux = flux.reshape(1,-1)
    ferr = numpy.ma.power(ivar, -0.5).reshape(1,-1)
    sres = sres.reshape(1,-1)

    #-------------------------------------------------------------------
    # Fit the stellar continuum

    # Mask the 5577 sky line and the emission lines
    sc_pixel_mask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                      emldb=EmissionLineDB.from_key('ELPSCMSK'))

    # Construct the template library
    sc_tpl = TemplateLibrary(sc_tpl_key, tpllib_list=tpllib_list, match_resolution=False,
                             velscale_ratio=velscale_ratio, spectral_step=1e-4, log=True,
                             hardcopy=False)
    sc_tpl_sres = numpy.mean(sc_tpl['SPECRES'].data, axis=0).ravel()

    # Instantiate the fitting class
    ppxf = PPXFFit(StellarContinuumModelBitMask())

    # Perform the fit
    cont_wave, cont_flux, cont_mask, cont_par \
        = ppxf.fit(sc_tpl['WAVE'].data.copy(), sc_tpl['FLUX'].data.copy(), wave, flux, ferr,
                   z, dispersion, iteration_mode='no_global_wrej', reject_boxcar=100,
                   ensemble=False, velscale_ratio=velscale_ratio, mask=sc_pixel_mask,
                   matched_resolution=False, tpl_sres=sc_tpl_sres, obj_sres=sres, degree=8,
                   moments=2, plot=False)

    # How many templates did it find:
    print(sc_tpl.ntpl)
    # From what files:
    print(sc_tpl.file_list)
    # What were the weights assigned to each template
    print(cont_par['TPLWGT'][0,:])

    # Remask the continuum fit
    sc_continuum = StellarContinuumModel.reset_continuum_mask_window(
                        numpy.ma.MaskedArray(cont_flux, mask=cont_mask>0))

    # Show the fit and residual
    pyplot.plot(wave, flux[0,:], label='Data')
    pyplot.plot(wave, sc_continuum[0,:], label='Model')
    pyplot.plot(wave, flux[0,:] - sc_continuum[0,:], label='Resid')
    pyplot.legend()
    pyplot.xlabel('Wavelength')
    pyplot.ylabel('Flux')
    pyplot.show()
    #-------------------------------------------------------------------

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




"""
WARNING:

    This is not a viable script, but is given as an example of what *could* be done.

    See https://sdss-mangadap.readthedocs.io/en/latest/development.html#adding-a-new-binning-scheme

"""

from pathlib import Path
import time
import warnings
import numpy
import astropy.constants

from mangadap.datacube.manga import MaNGADataCube
from mangadap.proc.reductionassessments import ReductionAssessment
from mangadap.proc.spectralstack import SpectralStackPar, SpectralStack
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectra, SpatiallyBinnedSpectraDef
from mangadap.proc.stellarcontinuummodel import StellarContinuumModel
from mangadap.proc.emissionlinemoments import EmissionLineMoments
from mangadap.proc.emissionlinemodel import EmissionLineModel
from mangadap.proc.spectralindices import SpectralIndices
from mangadap.dapfits import construct_maps_file, construct_cube_file

#-----------------------------------------------------------------------------

class ApertureBinning():
    """
    Perform aperture binning

    Args:
        x (array-like): List of on-sky x coordinates for apertures
        y (array-like): List of on-sky y coordinates for apertures
        r (array-like): Single or list of radii of the apertures

    Attributes:
        n (int): Number of apertures
        x (numpy.ndarray): On-sky x coordinates for apertures
        y (numpy.ndarray): On-sky y coordinates for apertures
        r (numpy.ndarray): Aperture radii
    """
    def __init__(self, x, y, r):
        self.x = numpy.asarray(x)
        if len(self.x.shape) != 1:
            raise ValueError('On-sky coordinates must be one-dimensional.')
        self.n = x.size
        if len(y) != self.n:
            raise ValueError('Input coordinates are of different lengths.')
        self.y = numpy.asarray(y)
        _r = numpy.atleast_1d(r)
        if len(_r) != 1 and len(_r) != self.n:
            raise ValueError('Radii must be common to all apertures or unique to each aperture.')
        self.r = numpy.full(self.n, r, dtype=float) if len(_r) == 1 else _r

    def bin_spaxels(self, x, y, par=None):
        _x = numpy.asarray(x)
        if len(_x.shape) != 1:
            raise ValueError('On-sky coordinates must be one-dimensional.')
        nspaxels = _x.size
        if len(y) != nspaxels:
            raise ValueError('Input coordinates are of different lengths.')
        _y = numpy.asarray(y)

        # Find which spaxels land in each aperture
        indx = numpy.square(_x[:,None]-self.x[None,:]) + numpy.square(_y[:,None]-self.y[None,:]) \
                    < numpy.square(self.r[None,:])
        if numpy.any(numpy.sum(indx, axis=1) > 1):
            warnings.warn('Spaxels found in multiple apertures!')

        # Return the aperture index that each spaxel is within,
        # isolating only one aperture per spaxel; spaxels not in any
        # aperture have a bin ID of -1
        binid = numpy.full((nspaxels, self.n), -1, dtype=int)
        binid[indx] = numpy.array([numpy.arange(self.n)]*nspaxels)[indx]
        return numpy.amax(binid, axis=1)

#-----------------------------------------------------------------------------
if __name__ == '__main__':
    t = time.perf_counter()

    # Set the plate, ifu, and initial velocity/redshift
    plate = 7443
    ifu = 6102
    #vel = 8675.5
    #nsa_redshift = vel/astropy.constants.c.to('km/s').value
    nsa_redshift = 0.0280161
    vel = nsa_redshift * astropy.constants.c.to('km/s').value

    input_path = Path('.').resolve()
    output_path = Path('.').resolve() / 'test_binning_output'

    # Read the DRP LOGCUBE file
    cube = MaNGADataCube.from_plateifu(plate, ifu, directory_path=input_path)

    # Calculate the S/N and coordinates
    rdxqa = ReductionAssessment('SNRG', cube, output_path=output_path)

    # Setup the aperture binning class
    ax = numpy.array([0.0, 3.0, 6.0])
    ay = numpy.array([0.0, 0.0, 0.0])
    apbin = ApertureBinning(ax, ay, 1.25)

    # Setup the stacking operations
    stackpar = SpectralStackPar(# Operation for stack
                                operation='mean',
                                # Apply a velocity registration
                                vel_register=False,
                                # Velocity offsets for registration
                                vel_offsets=None,
                                # Covariance mode and parameters
                                covar_mode='channels',
                                covar_par=SpectralStack.parse_covariance_parameters('channels', 11),
                                # Propagate the LSF through the stacking
                                stack_sres=True,
                                # Use pre-pixelized LSF (KHRR added this)
                                prepixel_sres=True)
    stacker = SpectralStack()

    # Create a new binning method
    binning_method = SpatiallyBinnedSpectraDef(# Key for binning method
                                               key='Aperture',
                                               # Galactic reddening function to use
                                               galactic_reddening='ODonnell',
                                               # Rv for Galactic reddening
                                               galactic_rv=3.1,             
                                               # Minimum S/N to include
                                               minimum_snr=0.0,
                                               # Binning function
                                               binfunc=apbin.bin_spaxels,
                                               # Object with stacking parameters
                                               stackpar=stackpar,
                                               # Stacking class instance
                                               stackclass=stacker,
                                               # Stacking function
                                               stackfunc=stacker.stack_DRPFits,
                                               # Type of LSF characterization to use
                                               spec_res='spaxel',
                                               # Use pre-pixelized LSF (KHRR added this)
                                               prepixel_sres=True)

    # Bin the spectra using the new binning method
    binned_spectra = SpatiallyBinnedSpectra('Aperture',     # Key for binning method
                                            cube,           # DRP data to bin
                                            rdxqa,          # Cube coordinates and S/N assessments
                                            method_list=binning_method, # Methods to select from
                                            output_path=output_path)

    # The rest of this is just a single execution of the remaining
    # analysis steps in
    # $MANGADAP_DIR/python/mangadap/survey/manga_dap.py , with some
    # simplifications

    stellar_continuum = StellarContinuumModel('GAU-MILESHC', binned_spectra, guess_vel=vel,
                                              guess_sig=100., output_path=output_path)

    emission_line_moments = EmissionLineMoments('EMOMM', binned_spectra,
                                                stellar_continuum=stellar_continuum,
                                                redshift=nsa_redshift, output_path=output_path)

    emission_line_model = EmissionLineModel('EFITM', binned_spectra,
                                            stellar_continuum=stellar_continuum,
                                            redshift=nsa_redshift, dispersion=100.0,
                                            output_path=output_path)
        
    spectral_indices = SpectralIndices('INDXEN', binned_spectra, redshift=nsa_redshift,
                                       stellar_continuum=stellar_continuum,
                                       emission_line_model=emission_line_model,
                                       output_path=output_path)

    construct_maps_file(cube, rdxqa=rdxqa, binned_spectra=binned_spectra,
                        stellar_continuum=stellar_continuum,
                        emission_line_moments=emission_line_moments,
                        emission_line_model=emission_line_model,
                        spectral_indices=spectral_indices, redshift=nsa_redshift,
                        output_path=output_path, overwrite=True)

    construct_cube_file(cube, rdxqa=rdxqa, binned_spectra=binned_spectra,
                        stellar_continuum=stellar_continuum,
                        emission_line_moments=emission_line_moments,
                        emission_line_model=emission_line_model,
                        spectral_indices=spectral_indices, output_path=output_path, overwrite=True)

    print('Elapsed time: {0} seconds'.format(time.perf_counter() - t))




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""

Defines a catch-all class for useful constants.  These are meant to be
values that are **not** available elsewhere, such as
`astropy.constants`_.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import numpy

class DAPConstants:
    r"""
    Defines the following set of constants:

    +-----------------------------+-------------------------------+
    | Attribute                   | Value                         |
    +=============================+===============================+
    | :attr:`sig2fwhm`            | :math:`\sqrt{8\ln(2)}`        |
    +-----------------------------+-------------------------------+
    | :attr:`rad2arcs`            | :math:`3600\frac{180}{\pi}`   |
    +-----------------------------+-------------------------------+
    | :attr:`sidereal_year`       | :math:`31558175.779`          |
    +-----------------------------+-------------------------------+
    """

    # Convert from sigma to FWHM: FWHM = sig2fwhm * sig
    sig2fwhm = numpy.sqrt(8.0 * numpy.log(2.0))
    r"""
    Conversion factor from the standard deviation, :math:`\sigma`, of
    a Gaussian to its full-width at half maximum (FWHM).
    """

    # Convert from radians to arcseconds: arcsec = rad2arcs * radians
    rad2arcs = 60*60*180/numpy.pi
    """
    Conversion factor from radians to to arcseconds.
    """


    # Length of one sidereal year in seconds ()
    sidereal_year = 31558175.779
    """
    Length of a sidereal year (1.0000385 Gregorian years) in seconds.
    """

    def __init__(self):
        self.sig2fwhm = constants.sig2fwhm
        self.rad2arcs = constants.rad2arcs
        self.sidereal_year = constants.sidereal_year





import pytest

import os

from IPython import embed

from mangadap.survey.drpcomplete import DRPComplete
from mangadap.util.parser import DefaultConfig
from mangadap.tests.util import remote_data_file, drp_test_version, requires_drpcomplete


@requires_drpcomplete
def test_write_cfg():
    ofile = 'test.ini'
    if os.path.isfile(ofile):
        # Remove existing files from failed tests
        os.remove(ofile)

    # Read the test DRPComplete file.
    # TODO: Temporarily override
    drp_test_version = 'v3_1_1'
    drpc = DRPComplete(drpver=drp_test_version, directory_path=remote_data_file(), readonly=True)

    # Write the base-level configuration file
    drpc.write_config(ofile, plate=7815, ifudesign=3702)

    # Read it and check the output
    cfg = DefaultConfig(ofile)
    assert cfg.getint('plate') == 7815, 'Plate number is wrong'
    assert cfg.getbool('log'), 'Should be selecting the logarithmically binned data'
    assert cfg.get('sres_ext') is None, 'Spectral resolution extension should be undefined'
    assert cfg.getfloat('z') == 2.9382300e-02, 'Bad redshift'

    # Try to write it again with overwrite set to False
    with pytest.raises(FileExistsError):
        drpc.write_config(ofile, plate=7815, ifudesign=3702, overwrite=False)

    # Set the spectral resolution flags
    drpc.write_config(ofile, plate=7815, ifudesign=3702, overwrite=True, sres_ext='SPECRES',
                      sres_fill=False)

    # Read it and check the output
    cfg = DefaultConfig(ofile)
    assert cfg.getint('plate') == 7815, 'Plate number is wrong'
    assert cfg.getbool('log'), 'Should be selecting the logarithmically binned data'
    assert cfg.get('sres_ext') == 'SPECRES', 'Spectral resolution extension incorrect'
    assert cfg.getfloat('z') == 2.9382300e-02, 'Bad redshift'

    # Clean-up
    os.remove(ofile)






from IPython import embed

from mangadap.par.absorptionindexdb import AbsorptionIndexDB


def test_read():
    dbs = AbsorptionIndexDB.available_databases()
    assert len(dbs) > 0, 'No absorption-line-index databases available'
    for key in dbs.keys():
        absdb = AbsorptionIndexDB.from_key(key)


def test_lick():
    absdb = AbsorptionIndexDB.from_key('LICKINDX')
    assert len(absdb) == 21, 'Incorrect number of Lick indices'
    assert 'Hb' in absdb['name'], 'Does not contain Hb in list'





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of processing utility functions for the MaNGA DAP.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import os
import glob
import warnings

import numpy

from scipy import interpolate, spatial
import astropy.constants

def HDUList_mask_wavelengths(hdu, bitmask, bitmask_flag, wave_limits, wave_ext='WAVE', \
                             mask_ext='MASK', invert=False):
    """
    Mask pixels in a specified wavelength range by turning on the bit
    value in the specified extention in a provided HDUList object.

    Args:
        hdu (`astropy.io.fits.HDUList`_): HDUList to alter
        bitmask (class:`BitMask`): Bit mask object used to turn on the
            named bit mask.
        bitmask_flag (str): Name of the bit to turn on.
        wave_limits (list or numpy.ndarray): Two-element array with the
            low and high wavelength limits.
        wave_ext (str): (Optional) Name of the wavelength extension in
            *hdu*.
        mask_ext (str): (Optional) Name of the mask extension in *hdu*.
        invert (bool): (Optional) Invert the sense of the masking.
            Instead of masking pixel in the wavelength interval, mask
            all pixels *outside* it.

    Returns:
        `astropy.io.fits.HDUList`_: The modified HDUList object.

    Raises:
        Exception: Raised if *wave_limits* does not have a length of
            two.

    """
    if len(wave_limits) != 2:
        raise Exception('Wavelength limits must be a two-element vector.')

    indx = numpy.where( (hdu[wave_ext].data < wave_limits[0]) \
                        | (hdu[wave_ext].data > wave_limits[1])) if invert else \
           numpy.where( (hdu[wave_ext].data >= wave_limits[0]) \
                        & (hdu[wave_ext].data <= wave_limits[1]))
    hdu[mask_ext].data[indx] = bitmask.turn_on(hdu[mask_ext].data[indx], bitmask_flag)
    return hdu


def select_proc_method(method_key, method_type, method_list=None, available_func=None):
    r"""
    Select a method from a list.  One of method_list or available_func
    must be provided.

    Args:
        method_key (:obj:`str`):
            Keyword used to select the method.
        method_type (object):
            Object type to check ``method_list`` against.
        method_list (:obj:`list`, optional):
            List of methods from which to find the selection keyword.
            If None, ``available_func`` **must** be provided.
        available_func (callable, optional):
            Callable function that returns a list of default methods
            in place of ``method_list``. For example, see
            :func:`mangadap.proc.templatelibrary.available_template_libraries`.

    Returns:
        object: An object with base class :class:`mangadap.par.ParSet`,
        containing a set of parameters used to define a method or
        database.

    Raises:
        KeyError:
            Raised if the selected keyword is not among the provided
            list or if the provided list has more than one identical
            keyword.
        TypeError:
            Raised if the input *method_list* object is not a list or
            *method_type*, or if available function is not a callable
            function.
    """
    # Get the default methods if no list provided
    if method_list is None:
        if not callable(available_func):
            raise TypeError('If not providing a list, must provide a callable function to ' \
                            'produce the default list of methods/databases/libraries.')
        method_list = available_func()

    # Make sure the methods have the right type
    if not isinstance(method_list, list):
        method_list = [method_list]
    for l in method_list:
        if not isinstance(l, method_type):
            raise TypeError('Input method/database/library must have type {0}')

    # Find the selected method via its keyword
    selected_method = [ l['key'] == method_key for l in method_list ]
    if numpy.sum(selected_method) == 0:
        raise KeyError('{0} is not a valid method/database/library!'.format(method_key))
    if numpy.sum(selected_method) > 1:
        for l in method_list:
            print('  ' + l['key'])
        print(method_key)
        raise KeyError('Keywords are not all unique!')

    # Return the method selected via the input keyword
    indx = numpy.where(selected_method)[0][0]
    return method_list[indx]


def get_database_key(f):
    """
    Construct a key from the provided file or file path.

    The key is a capitalized version of the file after removing any
    extension.

    Args:
        f (:obj:`str`):
            The file name or path.

    Returns:
        :obj:`str`: The keyword.

    Examples:

        >>> get_database_key('junk')
        'JUNK'
        >>> get_database_key('test.par')
        'TEST'
        >>> get_database_key('/path/to/test.par')
        'TEST'

    """
    return os.path.split(f)[1].split('.')[0].upper()

def select_database(key, directory_path):
    r"""
    Select a database using a keyword and directory path.

    Args:
        key (:obj:`str`):
            Keyword used to select the method.
        directory_path (:obj:`str`):
            Full path with the valid database files. All files in the
            directory with a ``.par`` extension will be included.

    Returns:
        :obj:`str`: Returns the file with the selected database.

    Raises:
        NotADirectoryError:
            Raised if the provided directory path does not exist.j
        KeyError:
            Raised if the selected keyword cannot be associated with
            a file in the provided directory.
    """
    if not os.path.isdir(directory_path):
        raise NotADirectoryError('{0} not found!'.format(directory_path))

    files = glob.glob(os.path.join(directory_path, '*.par'))
    keys = [get_database_key(f) for f in files]
    if key not in keys:
        raise KeyError('No database found to associate with {0}.'.format(key))

    return files[numpy.where(numpy.array(keys) == key)[0][0]]


#def _fill_vector(v, length, missing, value):
#    if v.size == length:
#        v[missing] = value
#        return v
#    _v = numpy.full(length, value, dtype=v.dtype)
#    _v[ list(set( numpy.arange(length) ) - set(missing)) ] = v
#    return _v


def flux_to_fnu(wave, flambda, unit_norm=1e-17):
    r"""

    Convert a spectrum with flux per unit wavelength to flux per unit
    frequency; i.e., calculate
    
    .. math::
        
        F_{\nu} = F_{\lambda} \frac{d\lambda}{d\nu} = F_{\lambda}
        \frac{\lambda^2}{c},

    where the first two arguments of the function are :math:`\lambda`
    and :math:`F_{\lambda}`.  The input wavelength units are expected to
    be angstroms, and the input flux units are expected to be :math:`n\
    {\rm erg\ s}^{-1}\ {\rm cm}^{-2}\ {\rm A}^{-1}`, where :math:`n` is the
    value of *unit_norm*.  The output flux units are microjanskys,
    :math:`10^{-29} {\rm erg\ s}^{-1}\ {\rm cm}^{-2}\ {\rm Hz}^{-1}`.

    Args:
        wave (numpy.ndarray, list): The vector with the wavelengths in
            angstroms.
        flambda (numpy.ndarray, list): The vector with the flux per unit
            wavelength (angstroms).
        unit_norm (float): (**Optional**) The unit normalization of the
            flux.  For example, this is :math:`10^{-17}` when the flux
            units are :math:`10^{-17} {\rm erg\ s}^{-1}\ {\rm cm}^{-2}\
            {\rm A}^{-1}`.

    Returns:
        float,numpy.ndarray: The flux in units of microjanskys.

    Raises:
        ValueError: Raised if the arguments do not have the same shape.
    """
    _wave = [wave] if isinstance(wave, float) else wave
    _wave = numpy.array(_wave) if isinstance(_wave, list) else _wave
    _flambda = [flambda] if isinstance(flambda, float) else flambda
    _flambda = numpy.array(_flambda) if isinstance(_flambda, list) else _flambda
    if _wave.shape != _flambda.shape:
        raise ValueError('Wavelength and flux arrays must have the same shape.')
    fnu = _flambda*numpy.square(_wave)*unit_norm*1e29/astropy.constants.c.to('angstrom/s').value
    return fnu[0] if isinstance(flambda, float) else fnu


        # TODO: Requires a spectrum in each bin!
#        if any(numpy.diff(bin_indx[srt]) > 1):
#            rep = numpy.ones(bin_change.size, dtype=int)
#            i = 1
#            while any(numpy.diff(bin_indx[srt]) > i):
#                rep[ numpy.where(numpy.diff(bin_indx[srt]) > i)[0]+1 == bin_change ] += 1
#                i += 1
#            bin_change = numpy.repeat(bin_change, rep)


# OLD VERSION
#def residual_growth(resid, growth_samples):
#    """
#    Interpolate the growth curve at distinct fractions, bracketed by the
#    minimum and maximum.
#    """
#    np = resid.size
#    grw = numpy.arange(np).astype(float)/np
#    resid_sort = numpy.sort(numpy.absolute(resid))
#    interp = interpolate.interp1d(grw, resid_sort, fill_value='extrapolate')
#    return numpy.append(numpy.append(resid_sort[0], interp(growth_samples)), resid_sort[-1])


#def residual_growth(resid, growth_samples):
#    """
#    Sample a set of residuals at specific the growth intervals.  No
#    interpolation is performed.
#    """
#    np = resid.size
#    resid_sort = numpy.sort(numpy.absolute(resid))
#    i = numpy.zeros(len(growth_samples)+2, dtype=float)
#    i[1:-1] = np*numpy.asarray(growth_samples)
#    i[-1] = np-1
#    i[i < 0] = 0
#    i[i >= np] = np-1
#    return resid_sort[i.astype(int)]


def sample_growth(a, samples, default=-9999., use_interpolate=True):
    _samples = numpy.asarray(samples)
    if numpy.any((_samples < 0) | (_samples > 1)):
        raise ValueError('Growth samples must be between 0 and 1.')
    _a = a.compressed() if isinstance(a, numpy.ma.MaskedArray) else numpy.atleast_1d(a).ravel()
    ns = _samples.size
    if len(_a) < 2:
        return [default]*ns if ns > 1 else default
    srt = numpy.argsort(_a)
    n = _a.size
    grw = (numpy.arange(n,dtype=float)+1)/n
    if use_interpolate:
        interpolator = interpolate.interp1d(grw, _a[srt], fill_value='extrapolate')
        g = interpolator(_samples)
        return tuple(g) if ns > 1 else g
    i = (n*_samples).astype(int)
    i[i > n-1] = n-1
    return _a[srt][i]


def growth_lim(a, lim, fac=1.0, midpoint=None, default=[0., 1.]):
    """
    Set the plots limits of an array based on two growth limits.

    Args:
        a (array-like): Array for which to determine limits.
        lim (float): Percentage of the array values to cover.
        fac (float): (**Optional**) Factor to increase the range based
            on the growth limits.  Default is no increase.
        midpoint (float): (**Optional**) Force the midpoint of the range
            to be centered on this value.  Default is to middle of
            growth range.
        default (list): (**Optional**) Default range to return if `a`
            has no data.  Default is 0 to 1.

    Returns:
        list: Lower and upper limits for the range of a plot of the data
        in `a`.
    """
    # Get the values to plot
    _a = a.compressed() if isinstance(a, numpy.ma.MaskedArray) else numpy.asarray(a).ravel()
    if len(_a) == 0:
        # No data so return the default range
        return default

    # Sort the values
    srt = numpy.ma.argsort(_a)

    # Set the starting and ending values based on a fraction of the
    # growth
    _lim = 1.0 if lim > 1.0 else lim
    start = int(len(_a)*(1.0-_lim)/2)
    end = int(len(_a)*(_lim + (1.0-_lim)/2))
    if end == len(_a):
        end -= 1

    # Set the full range and increase it by the provided factor
    Da = (_a[srt[end]] - _a[srt[start]])*fac

    # Set the midpoint if not provided
    mid = (_a[srt[start]] + _a[srt[end]])/2 if midpoint is None else midpoint

    # Return the range for the plotted data
    return [ mid - Da/2, mid + Da/2 ]


def optimal_scale(dat, mod, wgt=None):
    r"""
    Calculate the optimal scaling of an input model that minimizes the
    weighted root-mean-square difference between a set of data and a
    model.  When defining the weighted RMS as:

    .. math::
        
        {\rm RMS}^2 = \frac{1}{N} \sum_i w_i^2(d_i - f m_i)^2
    
    The optimal renormalization factor that minimizes the RMS is:

    .. math::
        
        f = \frac{\mathbf{d}^\prime \dot
            \mathbf{m}^\prime}{||\mathbf{m}^\prime||^2}

    where :math:`d^\prime_i = w_i d_i` and :math:`m^\prime_i = w_i m_i`.

    Args:
        dat (array-like): Array of data.
        mod (array-like): Model to renormalize.
        wgt (array-like): (**Optional**) Array of weights to apply to
            each residual.

    Returns:
        float : The optimal scaling that minimizes the weighted
        root-mean-square difference betwen the data and the model.

    Raises:
        ValueError: Raised if the array sizes do not match.
    """
    _dat = numpy.atleast_1d(dat)
    _mod = numpy.atleast_1d(mod)
    _wgt = numpy.ones(_dat.shape, dtype=float) if wgt is None else numpy.atleast_1d(wgt)
    if _mod.shape != _dat.shape or (_wgt is not None and _wgt.shape != _dat.shape):
        raise ValueError('Shapes of all input arrays must match.')

    dp = _dat*_wgt
    norm_mp = numpy.sum(numpy.square(_wgt*_mod))
    if norm_mp == 0:
        warnings.warn('Scale not determined because model norm is 0.')

    return 1.0 if norm_mp == 0 else numpy.sum(numpy.square(_wgt)*_dat*_mod)/norm_mp


def replace_with_data_from_nearest_coo(coo, data, replace):
    """
    Replace data in array with the spatially nearest neighbor.

    Args:
        coo (numpy.ndarray): A 2D array with the x and y coordinates of
            all the data.  Shape must be (NDATA, 2).

        data (numpy.ndarray): A 1D or 2D array with data to replace.
            The length of the first (or only) axis must be NDATA.

        replace (numpy.ndarray): Boolean array that is True for elements
            that should be replaced on output.  Shape must be (NDATA,).

    Returns:
            numpy.ndarray: The data array with the selected rows
            replaced with the nearest data set.

    Raises:
        ValueError: Raised if the array sizes are inappropriate.

    """
    # Check the input
    _coo = numpy.asarray(coo)
    if len(_coo.shape) != 2:
        raise ValueError('Input coordinate array must be two-dimensional.')
    if _coo.shape[1] != 2:
        raise ValueError('Currently only works with two coordinates.')
    ndata = _coo.shape[0]
    _data = numpy.asarray(data)
    if _data.shape[0] != ndata:
        raise ValueError('Coordinate and data arrays have a mismatched shape.')
    oned = len(_data.shape) == 1
    _replace = numpy.asarray(replace, dtype=bool)
    if len(_replace.shape) != 1 or _replace.shape[0] != ndata:
        raise ValueError('Input replacement selection array has an incorrect shape.')

    # Nothing flagged to replace, so just return a copy of the input
    if numpy.sum(_replace) == 0:
        return data.copy()

    # Use the coordinates to replace to set the KDTree reference grid
    do_not_replace = numpy.invert(_replace)
    kd = spatial.KDTree(_coo[do_not_replace,:])

    # Get the indices of the nearest data points
    dist, nearest_bin = kd.query(_coo[_replace,:])

    # Replace the existing data with the nearest one and return it
    new_data = _data.copy()
    if oned:
        new_data[_replace] = _data[do_not_replace][nearest_bin]
    else:
        new_data[_replace,:] = _data[do_not_replace,:][nearest_bin,:]
    return new_data


def inverse(d):
    """
    Return 1/d, where any division by 0 returns 0 instead of NaN or Inf.

    Args:
        d (scalar-like, array-like):
            Data values.

    Returns:
        float, `numpy.ndarray`_: Returns 1/d where values with d ==
        0. are replaced by 0. Return type matches input type: float
        for scalar, `numpy.ndarray`_ for array-like.
    """
    _d = float(d) if isinstance(d, (float, int)) else numpy.atleast_1d(d).astype(float)
    m = _d != 0.0
    return m/(_d + numpy.logical_not(m))

